
import SwiftUI
import MetalKit

struct MetalView: View {
    @Environment(\.colorScheme) var colorScheme
    
    @State private var metalView = MTKView()
    @State private var renderer: Renderer?
    @State private var doubleTapTime = 0.0
    @State private var showSettings = true
    @State private var isImporting = false
    
    @State private var lineSizeControl = 1.0
    @State private var ballSizeControl = 1.0
    @State private var trackballControl = false
    @State private var centerObjectControl = false
    @State private var upDirectionControl = "Z"
    @State private var invertObjectControl = false
    @State private var handednessControl = false
    @State private var stereoModeControl = StereoMode.mono
    @State private var stereoAngleControl = 0.0
    @State private var backgroundColorControl = 0
    @State private var wireFrameControl = false
    
    private var stereoAngleDisplay: String { String(round(stereoAngleControl * 10.0) / 10.0) }
    
    private let lightHaptic = UIImpactFeedbackGenerator(style: .light)
    private let mediumHaptic = UIImpactFeedbackGenerator(style: .medium)
    
    class SharedVarsStorage: ObservableObject { @Published var fps: String = "" }
    @StateObject var sharedVars = SharedVarsStorage()
    func updateFPS(_ updatedText: String) { sharedVars.fps = updatedText }
    
    init() {
        // Set color of active controls
        UISegmentedControl.appearance().selectedSegmentTintColor = UIColor(red: 0.6, green: 0.8, blue: 1.0, alpha: 0.5)
    }
    
    
    // ###################################################################################################
    // VIEW BODY
    // ###################################################################################################
    
    var body: some View {
        // Rotating layout - settings on bottom for portrait and right side for landscape
        let layout = UIDevice.current.orientation == .landscapeLeft || UIDevice.current.orientation == .landscapeRight ? AnyLayout(HStackLayout(spacing: 0)) : AnyLayout(VStackLayout(spacing: 0))
        
        layout {
            MetalViewRepresentable(metalView: $metalView)
            
            // ###### Create renderer and set control values ######
                .onAppear {
                    renderer = Renderer(metalView: metalView, inputText: demo1, updateFPS: updateFPS)
                    lineSizeControl = pow(renderer!.lineSize*40000, 1.0/3.0)
                    ballSizeControl = pow(renderer!.ballSize*40000, 1.0/3.0)
                    trackballControl = renderer!.trackball
                    centerObjectControl = renderer!.centerObject
                    upDirectionControl = String(renderer!.upDirection)
                    invertObjectControl = renderer!.invertObject
                    handednessControl = renderer!.handedness
                    stereoModeControl = renderer!.stereoMode
                    stereoAngleControl = renderer!.stereoAngle
                    backgroundColorControl = renderer!.backgroundColor
                    wireFrameControl = renderer!.wireFrame
                }
            
            // ###### Rotate Object ######
                .gesture(
                    DragGesture(minimumDistance: 0.0, coordinateSpace: .global)
                        .onChanged { gesture in
                            let thepoint = CGPoint(x: gesture.translation.width, y: gesture.translation.height)
                            renderer!.dragUpdate(currentMousePoint: thepoint, dragEnded: false, metalView: metalView)
                        }
                        .onEnded { _ in
                            let thepoint = CGPoint(x: 0, y: 0)
                            renderer!.dragUpdate(currentMousePoint: thepoint, dragEnded: true, metalView: metalView)
                            let currentTime = CFAbsoluteTimeGetCurrent()
                            if currentTime - doubleTapTime < 0.25 { showSettings = !showSettings }
                            doubleTapTime = currentTime
                        }
                )
            
            // ###### Zoom Object ###### for 17.0+
            //    .gesture (
            //        MagnifyGesture()
            //          .onChanged { gesture in
            //                let themag = gesture.magnification
            //                renderer!.zoomUpdate(zoomChange: themag, zoomEnded: false, metalView: metalView)
            //            }
            //            .onEnded { gesture in
            //                let themag = gesture.magnification
            //                renderer!.zoomUpdate(zoomChange: -themag, zoomEnded: true, metalView: metalView)
            //            }
            //    )
            
            // ###### Zoom Object ###### for 16.0+
                .gesture (
                    MagnificationGesture()
                        .onChanged { gesture in
                            let themag = gesture
                            renderer!.zoomUpdate(zoomChange: themag, zoomEnded: false, metalView: metalView)
                        }
                        .onEnded { gesture in
                            let themag = gesture
                            renderer!.zoomUpdate(zoomChange: themag, zoomEnded: true, metalView: metalView)
                        }
                )
            
            
            // ###################################################################################################
            // SETTINGS
            // ###################################################################################################
            
            if showSettings {
                Divider() .overlay(colorScheme == .light ? Color.black : Color.white)
                ScrollView {
                    VStack {
                        Text("").frame(maxWidth: .infinity)
                        
                        // ###### Top Line Buttons - Close / FPS / Open ######
                        HStack {
                            // ### Close Settings ###
                            Button { showSettings = false } label: { Text(String("Close")) .bold() }
                                .frame(maxWidth: .infinity, alignment: .leading)
                            
                            
                            // ### FPS Display ###
                            Text(sharedVars.fps)
                                .frame(maxWidth: .infinity, alignment: .center)
                                .font(.body)
                                .bold()
                                .foregroundStyle(.red)
                            
                            
                            // ### Open File ###
                            Button { isImporting = true } label: { Text(String("Open")) .bold() }
                                .fileImporter( isPresented: $isImporting, allowedContentTypes: [.data], allowsMultipleSelection: false ) { result in
                                    do {
                                        guard let selectedFile: URL = try result.get().first
                                        else { print("No File"); return }
                                        if selectedFile.startAccessingSecurityScopedResource() {
                                            guard let input = String(data: try Data(contentsOf: selectedFile), encoding: .isoLatin1)
                                            else { print("No Input"); return }
                                            defer { selectedFile.stopAccessingSecurityScopedResource() }
                                            let fileText = input
                                            renderer!.readDataText(inputText: fileText, metalView: metalView)
                                            return
                                        } else {
                                            // Handle denied access
                                            print("Denied access")
                                            return
                                        }
                                    } catch {
                                        // Handle failure
                                        print("Unable to read file contents")
                                        //print(error.localizedDescription)
                                        return
                                    }
                                }
                                .frame(maxWidth: .infinity, alignment: .trailing)
                        }
                        .frame(width: 300)
                        
                        
                        // ######################################################
                        Spacer(minLength: 10)
                        Divider() .frame(width: 300) .overlay(colorScheme == .light ? Color.black : Color.white)
                        Spacer(minLength: 15)
                        // ######################################################
                        
                        
                        // ###### Line Size ######
                        HStack {
                            Text("Line Size")
                                .frame(maxWidth: .infinity, alignment: .leading)
                                .font(.subheadline)
                            Slider(value: $lineSizeControl, in: 1...40)
                                .onChange(of: lineSizeControl) { newValue in
                                    renderer!.lineSize = lineSizeControl*lineSizeControl*lineSizeControl/40000.0
                                    renderer!.makeScene(metalView: metalView)
                                    metalView.isPaused = false
                                    lightHaptic.impactOccurred()
                                }
                                .frame(width: 200)
                                .frame(maxWidth: .infinity, alignment: .trailing)
                        }
                        .frame(width: 300)
                        
                        
                        // ###### Dot Size ######
                        HStack () {
                            Text("Dot Size")
                                .frame(maxWidth: .infinity, alignment: .leading)
                                .font(.subheadline)
                            Slider(value: $ballSizeControl, in: 1...40)
                                .onChange(of: ballSizeControl) { newValue in
                                    renderer!.ballSize = ballSizeControl*ballSizeControl*ballSizeControl/40000.0
                                    renderer!.makeScene(metalView: metalView)
                                    metalView.isPaused = false
                                    lightHaptic.impactOccurred()
                                }
                                .frame(width: 200)
                                .frame(maxWidth: .infinity, alignment: .trailing)
                        }
                        .frame(width: 300)
                        
                        
                        // ######################################################
                        Spacer(minLength: 15)
                        Divider() .frame(width: 300) .overlay(colorScheme == .light ? Color.black : Color.white)
                        Spacer(minLength: 15)
                        // ######################################################
                        
                        
                        // ###### Turntable / Trackball ######
                        Picker("Trackball", selection: $trackballControl) {
                            Text("Turntable").tag(false)
                            Text("Trackball").tag(true)
                        }
                        .pickerStyle(.segmented)
                        .onChange(of: trackballControl) { newValue in
                            renderer!.trackball = trackballControl
                            renderer!.makeScene(metalView: metalView)
                            metalView.isPaused = true
                            renderer!.updateNodeConstants()
                            metalView.isPaused = false
                            mediumHaptic.impactOccurred()
                        }
                        .frame(width: 300)
                        
                        
                        // ###### Absolute / Centered ######
                        Picker("Center Object", selection: $centerObjectControl) {
                            Text("Absolute").tag(false)
                            Text("Centered").tag(true)
                        }
                        .pickerStyle(.segmented)
                        .onChange(of: centerObjectControl) { newValue in
                            renderer!.centerObject = centerObjectControl
                            renderer!.makeScene(metalView: metalView)
                            metalView.isPaused = false
                            mediumHaptic.impactOccurred()
                        }
                        .frame(width: 300)
                        
                        
                        // ###### Z Y X Up ######
                        Picker("Up Direction", selection: $upDirectionControl) {
                            Text("Z is Up").tag("Z")
                            Text("Y is Up").tag("Y")
                            Text("X is Up").tag("X")
                        }
                        .pickerStyle(.segmented)
                        .onChange(of: upDirectionControl) { newValue in
                            renderer!.upDirection = Character(upDirectionControl)
                            renderer!.makeScene(metalView: metalView)
                            metalView.isPaused = false
                            mediumHaptic.impactOccurred()
                        }
                        .frame(width: 300)
                        
                        
                        // ###### Upright / Inverted ######
                        Picker("Inverting", selection: $invertObjectControl) {
                            Text("Upright").tag(false)
                            Text("Inverted").tag(true)
                        }
                        .pickerStyle(.segmented)
                        .onChange(of: invertObjectControl) { newValue in
                            renderer!.invertObject = invertObjectControl
                            renderer!.makeScene(metalView: metalView)
                            metalView.isPaused = false
                            mediumHaptic.impactOccurred()
                        }
                        .frame(width: 300)
                        
                        
                        // ###### Right Handed / Left Handed ######
                        Picker("Handedness", selection: $handednessControl) {
                            Text("Right Handed").tag(false)
                            Text("Left Handed").tag(true)
                        }
                        .pickerStyle(.segmented)
                        .onChange(of: handednessControl) { newValue in
                            renderer!.handedness = handednessControl
                            renderer!.makeScene(metalView: metalView)
                            metalView.isPaused = false
                            mediumHaptic.impactOccurred()
                        }
                        .frame(width: 300)
                        
                        
                        // ######################################################
                        Spacer(minLength: 15)
                        Divider() .frame(width: 300) .overlay(colorScheme == .light ? Color.black : Color.white)
                        Spacer(minLength: 15)
                        // ######################################################
                        
                        
                        // ###### Mono / Stereo mode ######
                        Picker("Stereo Mode", selection: $stereoModeControl) {
                            Text("Mono").tag(StereoMode.mono)
                            Text("Dual").tag(StereoMode.dual)
                            Text("RG").tag(StereoMode.RG)
                            Text("RB").tag(StereoMode.RB)
                        }
                        .pickerStyle(.segmented)
                        .onChange(of: stereoModeControl) { newValue in
                            renderer!.stereoMode = stereoModeControl
                            renderer!.makeScene(metalView: metalView)
                            metalView.isPaused = false
                            mediumHaptic.impactOccurred()
                        }
                        .frame(width: 300)
                        
                        
                        // ###### Stereo Angle ######
                        HStack {
                            Text("Angle (\(stereoAngleDisplay)°)")
                                .frame(maxWidth: .infinity, alignment: .leading)
                                .font(.subheadline)
                            Slider(value: $stereoAngleControl, in: -6...6, step: 0.1)
                                .onChange(of: stereoAngleControl) { newValue in
                                    lightHaptic.impactOccurred()
                                    renderer!.stereoAngle = round(stereoAngleControl * 10.0) / 10.0
                                    metalView.isPaused = false
                                }
                                .frame(width: 180)
                                .frame(maxWidth: .infinity, alignment: .trailing)
                        }
                        .frame(width: 300)
                        
                        
                        // ######################################################
                        Spacer(minLength: 15)
                        Divider() .frame(width: 300) .overlay(colorScheme == .light ? Color.black : Color.white)
                        Spacer(minLength: 15)
                        // ######################################################
                        
                        
                        // ###### Background Color ######
                        Picker("Background", selection: $backgroundColorControl) {
                            Text("Black").tag(0)
                            Text("Gray").tag(2)
                            Text("White").tag(1)
                        }
                        .pickerStyle(.segmented)
                        .onChange(of: backgroundColorControl) { newValue in
                            renderer!.backgroundColor = backgroundColorControl
                            renderer!.setBackgroundColor(backgroundColorControl, metalView: metalView)
                            renderer!.makeScene(metalView: metalView)
                            metalView.isPaused = false
                            mediumHaptic.impactOccurred()
                        }
                        .frame(width: 300)
                        
                        
                        // ###### Rendered / Wireframe ######
                        Picker("Rendering type", selection: $wireFrameControl) {
                            Text("Rendered").tag(false)
                            Text("Wireframe").tag(true)
                        }
                        .pickerStyle(.segmented)
                        .onChange(of: wireFrameControl) { newValue in
                            renderer!.wireFrame = wireFrameControl
                            renderer!.makeScene(metalView: metalView)
                            metalView.isPaused = false
                            mediumHaptic.impactOccurred()
                        }
                        .frame(width: 300)
                        
                        
                        // ######################################################
                        Spacer(minLength: 15)
                        Divider() .frame(width: 300) .overlay(colorScheme == .light ? Color.black : Color.white)
                        Spacer(minLength: 15)
                        // ######################################################
                        
                        
                        HStack {
                            // ### Demo 1 ###
                            Button { renderer!.readDataText(inputText: demo1, metalView: metalView) } label: { Text(String("Demo 1")) .bold() }
                                .frame(maxWidth: .infinity, alignment: .leading)
                            
                            
                            // ### Demo 2 ###
                            Button { renderer!.readDataText(inputText: demo2, metalView: metalView) } label: { Text(String("Demo 2")) .bold() }
                                .frame(maxWidth: .infinity, alignment: .center)
                            
                            
                            // ### Demo 3 ###
                            Button { renderer!.readDataText(inputText: demo3, metalView: metalView) } label: { Text(String("Demo 3")) .bold() }
                                .frame(maxWidth: .infinity, alignment: .trailing)
                        }
                        .frame(width: 300)
                        
                        
                        // ######################################################
                        Spacer(minLength: 15)
                        Divider() .frame(width: 300) .overlay(colorScheme == .light ? Color.black : Color.white)
                        Spacer(minLength: 15)
                        // ######################################################
                        
                        
                        // ###### Reset Rotation Angles ######
                        Button {
                            renderer!.resetAngles(metalView: metalView)
                            mediumHaptic.impactOccurred()
                        } label: { Text(String("Reset Rotation Angles")) .bold() }
                            .frame(width: 300)
                        
                        
                        // ######################################################
                        Spacer(minLength: 15)
                        // ######################################################
                        
                        
                        // ###### Factory Reset ######
                        Button {
                            renderer!.factoryReset(metalView: metalView)
                            // Read back settings for controls
                            lineSizeControl = pow(renderer!.lineSize*40000, 1.0/3.0)
                            ballSizeControl = pow(renderer!.ballSize*40000, 1.0/3.0)
                            trackballControl = renderer!.trackball
                            centerObjectControl = renderer!.centerObject
                            upDirectionControl = String(renderer!.upDirection)
                            invertObjectControl = renderer!.invertObject
                            handednessControl = renderer!.handedness
                            stereoModeControl = renderer!.stereoMode
                            stereoAngleControl = renderer!.stereoAngle
                            backgroundColorControl = renderer!.backgroundColor
                            wireFrameControl = renderer!.wireFrame
                            mediumHaptic.impactOccurred()
                        } label: { Text(String("Factory Reset")) .bold() }
                            .frame(width: 300)
                        
                        
                        // ######################################################
                        Spacer(minLength: 15)
                        // ######################################################
                        
                        
                        // ###### Save Defaults ######
                        Button {
                            renderer!.saveDefaults(metalView: metalView)
                            mediumHaptic.impactOccurred()
                        } label: { Text(String("Save Defaults")) .bold() }
                            .frame(width: 300)
                        
                        
                        // ######################################################
                        Spacer(minLength: 15)
                        Divider() .frame(width: 300) .overlay(colorScheme == .light ? Color.black : Color.white)
                        Spacer(minLength: 5)
                        // ######################################################
                        
                        
                        Text("Double Tap to reopen these settings") .font(.footnote)
                        
                        
                        // ######################################################
                        Spacer(minLength: 5)
                        Divider() .frame(width: 300) .overlay(colorScheme == .light ? Color.black : Color.white)
                        Spacer(minLength: 30)
                        // ######################################################
                    }
                }
                .padding(0)
                .background(colorScheme == .light ? Color(red: 0.9, green: 0.9, blue: 0.9) : nil )
                .frame(width: UIDevice.current.orientation == .landscapeLeft || UIDevice.current.orientation == .landscapeRight ? 350 : nil)
            } // Settings
            
        } // layout
        .edgesIgnoringSafeArea(.all)
        .persistentSystemOverlays(.hidden)
        .statusBar(hidden: true)
        .background(colorScheme == .light ? Color(red: 0.9, green: 0.9, blue: 0.9) : nil )
    }
}


struct MetalViewRepresentable: UIViewRepresentable {
    @Binding var metalView: MTKView
    func makeUIView(context: Context) -> MTKView { metalView }
    func updateUIView(_ uiView: MTKView, context: Context) { updateMetalView() }
    func updateMetalView() { }
}
