
import SwiftUI
import MetalKit

let MaxOutstandingFrameCount = 3 // Triple buffer the render
let MinBufferAlignment = 256

struct NodeConstants {
    var modelViewMatrix: float4x4
    var modelViewMatrix1: float4x4
    var color: Int
}

struct FrameConstants {
    var projectionMatrix: float4x4
    var lightCount: UInt32
    var stereoAngle: Float
    var stereoType: Int
}

struct LightConstants {
    var intensity: simd_float3
    var direction: simd_float3
    var type: UInt32
}

struct QuadTextureVertex {
    var position: vector_float2
    var texcoord: vector_float2
}

struct MoveHistory {
    var time: CFAbsoluteTime
    var x: Double
    var y: Double
}

enum StereoMode: Int {
    case mono = 0
    case dual = 1
    case RG = 2
    case RB = 3
}


//###########################################################################################//
//### RENDERER
//###########################################################################################//

class Renderer: NSObject {
    static var device: MTLDevice!
    static var commandQueue: MTLCommandQueue!
    
    var ballSize: Double
    var lineSize: Double
    var upDirection: Character
    var invertObject: Bool
    var handedness: Bool
    var centerObject: Bool
    var backgroundColor: Int
    var stereoMode: StereoMode
    var stereoAngle: Double
    var trackball: Bool
    var wireFrame: Bool
    
    private var vertexDescriptor: MTLVertexDescriptor!
    private var offScreenTexture: MTLTexture!
    private var offScreenDepthTexture: MTLTexture!
    private var MaxConstantsSize: Int = 0
    private var currentWidth: Double = 0.0
    private var currentHeight: Double = 0.0
    private var numSegmentsBall: UInt32 = 32
    private var numSegmentsLine: UInt32 = 32
    private var stereoModePointer: Int = 0
    private var quadVertices: [QuadTextureVertex]
    private var fpsCount: Int = 0
    private var fpsTime: CFAbsoluteTime = CFAbsoluteTimeGetCurrent()
    private var mouseTime: CFAbsoluteTime = CFAbsoluteTimeGetCurrent()
    private var frameTime: CFAbsoluteTime = CFAbsoluteTimeGetCurrent()
    private var previousMousePoint: CGPoint = CGPoint(x: 0, y: 0)
    private var cursorDeltaX: Double = 0.0
    private var cursorDeltaZ: Double = 0.0
    private var cursorDeltaZprevious: Double = 0.0
    private var cursorDeltaXprevious: Double = 0.25
    private var absoluteAngleX: Double = 0.0
    private var absoluteAngleZ: Double = 0.0
    private var zoomLevel: Double = 7.0
    private var zoomFactor: Double = 1.0
    private var trackballCurrent: simd_float4x4 = simd_float4x4(SIMD4<Float>(1, 0, 0, 0), SIMD4<Float>(0, 1, 0, 0), SIMD4<Float>(0, 0, 1, 0), SIMD4<Float>(0, 0, 0, 1))

    private var nodes = [Node]()
    private var lights = [Light]()
    private var objectDataArray: [[Double]] = []
    private var moveHistory: [MoveHistory] = []

    private var updateFPS: (String) -> ()
    
    private var constantBuffer: MTLBuffer!
    private var currentConstantBufferOffset = 0
    private var frameConstantsOffset = 0
    private var lightConstantsOffset = 0
    private var nodeConstantsOffsets = [Int]()
    private var renderPipelineState: MTLRenderPipelineState!
    private var renderPipelineState2: MTLRenderPipelineState!
    private var depthStencilState: MTLDepthStencilState!
    private var frameSemaphore = DispatchSemaphore(value: MaxOutstandingFrameCount)
    
    
    // Initialise Renderer
    init(metalView: MTKView, inputText: String, updateFPS: @escaping (_ value: String) -> ()) {
        // FPS update function feeding back into MetalView
        self.updateFPS = updateFPS
        
        // Read in saved defaults
        @AppStorage("DefaultBallSize") var DefaultBallSize = 0.04
        @AppStorage("DefaultLineSize") var DefaultLineSize = 0.02
        @AppStorage("DefaultUpDirection") var DefaultUpDirection = "Z"
        @AppStorage("DefaultInvertObject") var DefaultInvertObject = false
        @AppStorage("DefaultHandedness") var DefaultHandedness = false
        @AppStorage("DefaultCenterObject") var DefaultCenterObject = false
        @AppStorage("DefaultBackgroundColor") var DefaultBackgroundColor = 0
        @AppStorage("DefaultStereoMode") var DefaultStereoMode = 0
        @AppStorage("DefaultStereoAngle") var DefaultStereoAngle = 3.2
        @AppStorage("DefaultTrackball") var DefaultTrackball = false
        @AppStorage("DefaultWireFrame") var DefaultWireFrame = false
        ballSize = DefaultBallSize
        lineSize = DefaultLineSize
        upDirection = Character(DefaultUpDirection)
        invertObject = DefaultInvertObject
        handedness = DefaultHandedness
        centerObject = DefaultCenterObject
        backgroundColor = DefaultBackgroundColor
        stereoMode = StereoMode(rawValue: DefaultStereoMode)!
        stereoAngle = DefaultStereoAngle
        trackball = DefaultTrackball
        wireFrame = DefaultWireFrame
        
        quadVertices = [
            QuadTextureVertex(position: [ 1, -1], texcoord: [1, 1]),
            QuadTextureVertex(position: [-1, -1], texcoord: [0, 1]),
            QuadTextureVertex(position: [-1,  1], texcoord: [0, 0]),
            QuadTextureVertex(position: [ 1, -1], texcoord: [1, 1]),
            QuadTextureVertex(position: [-1,  1], texcoord: [0, 0]),
            QuadTextureVertex(position: [ 1,  1], texcoord: [1, 0]) ]
        
        super.init()
        
        guard let device = MTLCreateSystemDefaultDevice(), let commandQueue = device.makeCommandQueue()
        else { fatalError("GPU not available") }
        
        Renderer.device = device
        Renderer.commandQueue = commandQueue
        metalView.device = device
        metalView.framebufferOnly = true // false for accessible texture
        metalView.delegate = self
        metalView.preferredFramesPerSecond = 120
        metalView.colorPixelFormat = .bgra8Unorm_srgb
        metalView.depthStencilPixelFormat = .depth32Float
        metalView.isPaused = true
        metalView.presentsWithTransaction = false
        metalView.enableSetNeedsDisplay = true
        
        setBackgroundColor(backgroundColor, metalView: metalView)
        readDataText(inputText: inputText, metalView: metalView)
        makePipeline(metalView: metalView)
    }
}


extension Renderer: MTKViewDelegate {
    
    //###########################################################################################//
    //### MAKE THE SCENE
    //###########################################################################################//
    
    func makeScene(metalView: MTKView) {
        nodes = [Node]()
        
        let windowheight = metalView.bounds.height
        let windowzoom = 1.0 / (zoomLevel * zoomFactor)
        
        // More segments for larger on screen spheres
        numSegmentsBall = UInt32(sqrt(windowheight * windowzoom * ballSize) * 8)
        if numSegmentsBall < 8 { numSegmentsBall = 8}
        if numSegmentsBall > 64 { numSegmentsBall = 64}
        
        // Make multiple of 4 so cylinder and end cap facets line up
        numSegmentsLine = UInt32(sqrt(windowheight * windowzoom * lineSize) * 8 / 4)
        numSegmentsLine *= 4
        if numSegmentsLine < 8 { numSegmentsLine = 8}
        if numSegmentsLine > 64 { numSegmentsLine = 64}
        
        let allocator = MTKMeshBufferAllocator(device: Renderer.device)
        
        let mdlVertexDescriptor = MDLVertexDescriptor()
        mdlVertexDescriptor.vertexAttributes[0].name = MDLVertexAttributePosition
        mdlVertexDescriptor.vertexAttributes[0].format = .float3
        mdlVertexDescriptor.vertexAttributes[0].offset = 0
        mdlVertexDescriptor.vertexAttributes[0].bufferIndex = 0
        mdlVertexDescriptor.vertexAttributes[1].name = MDLVertexAttributeNormal
        mdlVertexDescriptor.vertexAttributes[1].format = .float3
        mdlVertexDescriptor.vertexAttributes[1].offset = 12 // 16 if 12 does not work
        mdlVertexDescriptor.vertexAttributes[1].bufferIndex = 0
        mdlVertexDescriptor.bufferLayouts[0].stride = 24 // 32 if 24 does not work
        vertexDescriptor = MTKMetalVertexDescriptorFromModelIO(mdlVertexDescriptor)!
        
        
        // Construct Ball Mesh
        var mdlBall = MDLMesh(sphereWithExtent: SIMD3<Float>(1.0, 1.0, 1.0), segments: SIMD2<UInt32>(numSegmentsBall, numSegmentsBall), inwardNormals: false, geometryType: .triangles, allocator: allocator)
        if wireFrame {
            mdlBall = MDLMesh(sphereWithExtent: SIMD3<Float>(1.0, 1.0, 1.0), segments: SIMD2<UInt32>(numSegmentsBall, numSegmentsBall), inwardNormals: false, geometryType: .lines, allocator: allocator)
        }
        mdlBall.vertexDescriptor = mdlVertexDescriptor
        let BallMesh = try! MTKMesh(mesh: mdlBall, device: Renderer.device)
        
        
        // Construct Line End Cap Mesh
        var mdlLineBall = MDLMesh(hemisphereWithExtent: SIMD3<Float>(1, 1, 1), segments: SIMD2<UInt32>(numSegmentsLine, 24), inwardNormals: false, cap: false, geometryType: .triangles, allocator: allocator)
        if wireFrame {
            mdlLineBall = MDLMesh(hemisphereWithExtent: SIMD3<Float>(1, 1, 1), segments: SIMD2<UInt32>(numSegmentsLine, 24), inwardNormals: false, cap: false, geometryType: .lines, allocator: allocator)
        }
        mdlLineBall.vertexDescriptor = mdlVertexDescriptor
        let LineBallMesh = try! MTKMesh(mesh: mdlLineBall, device: Renderer.device)
        
        
        // Construct Line Mesh
        var mdlLine = MDLMesh(cylinderWithExtent: SIMD3<Float>(1, 1, 1), segments: [numSegmentsLine, 1], inwardNormals: false, topCap: false, bottomCap: false, geometryType: .triangles, allocator: allocator)
        if wireFrame {
            mdlLine = MDLMesh(cylinderWithExtent: SIMD3<Float>(1, 1, 1), segments: [numSegmentsLine, 1], inwardNormals: false, topCap: false, bottomCap: false, geometryType: .lines, allocator: allocator)
        }
        mdlLine.vertexDescriptor = mdlVertexDescriptor
        let lineMesh = try! MTKMesh(mesh: mdlLine, device: Renderer.device)
        
        
        // Compute offsets for centering if needed
        var offsetX = 0.0
        var offsetY = 0.0
        var offsetZ = 0.0
        if centerObject {
            var minX: Double = .infinity
            var minY: Double = .infinity
            var minZ: Double = .infinity
            var maxX: Double = -.infinity
            var maxY: Double = -.infinity
            var maxZ: Double = -.infinity
            for a in objectDataArray {
                if a[0] < minX { minX = a[0]}
                if a[1] < minY { minY = a[1]}
                if a[2] < minZ { minZ = a[2]}
                if a[0] > maxX { maxX = a[0]}
                if a[1] > maxY { maxY = a[1]}
                if a[2] > maxZ { maxZ = a[2]}
            }
            offsetX = (maxX + minX) / 2
            offsetY = (maxY + minY) / 2
            offsetZ = (maxZ + minZ) / 2
        }
        
        // Get maximum node distance for scaling
        var max_dist = 0.0
        for a in objectDataArray {
            let x = a[0] - offsetX
            let y = a[1] - offsetY
            let z = a[2] - offsetZ
            let this_dist = sqrt(x*x + y*y + z*z)
            if this_dist > max_dist { max_dist = this_dist }
        }
        if max_dist == 0 { max_dist = 1 }
        
        
        // Turn Data Array into Nodes
        var last_point:[Double] = [-offsetX, -offsetY, -offsetZ, 0]
        
        for next_point in objectDataArray {
            var x0 = (last_point[0] - offsetX) / max_dist
            var y0 = (last_point[1] - offsetY) / max_dist
            var z0 = (last_point[2] - offsetZ) / max_dist
            var x1 = (next_point[0] - offsetX) / max_dist
            var y1 = (next_point[1] - offsetY) / max_dist
            var z1 = (next_point[2] - offsetZ) / max_dist
            let cc = Int(round(next_point[3]))
            var tt: Double
            
            // Default upDirection is Y
            if upDirection == "Z" { tt = y0; y0 = z0; z0 = tt; tt = y1; y1 = z1; z1 = tt; }
            if upDirection == "X" { tt = x0; x0 = y0; y0 = tt; tt = x1; x1 = y1; y1 = tt;}
            if invertObject { y0 = -y0; y1 = -y1; x0 = -x0; x1 = -x1; }
            if handedness { z0 = -z0; z1 = -z1 }
            
            // Show a Ball
            if cc < 0 {
                let ballNode = Node(mesh: BallMesh)
                ballNode.transform = simd_float4x4(translate: SIMD3<Float>(Float(x1), Float(y1), Float(z1))) * simd_float4x4(scale: SIMD3<Float>(Float(ballSize), Float(ballSize), Float(ballSize)))
                ballNode.color = -cc
                nodes.append(ballNode)
            }
            
            // Show a Line
            else if cc > 0 {
                var theaxis: SIMD3<Float>
                let length = sqrt((x1-x0)*(x1-x0) + (y1-y0)*(y1-y0) + (z1-z0)*(z1-z0))
                let a1 = atan2(y1-y0, x1-x0)
                let a2 = atan2(sqrt((x1-x0)*(x1-x0) + (y1-y0)*(y1-y0)), z1-z0)
                
                let lineNode = Node(mesh: lineMesh)
                lineNode.transform = lineNode.transform * simd_float4x4(translate: SIMD3<Float>(Float((x1+x0))/2, Float((y1+y0))/2, Float((z1+z0))/2))
                theaxis = SIMD3<Float>(1, 0, 0)
                lineNode.transform = lineNode.transform * simd_float4x4(rotateAbout: theaxis, byAngle: .pi/2)
                theaxis = SIMD3<Float>(0, 1, 0)
                lineNode.transform = lineNode.transform * simd_float4x4(rotateAbout: theaxis, byAngle: Float(a1))
                theaxis = SIMD3<Float>(0, 0, 1)
                lineNode.transform = lineNode.transform * simd_float4x4(rotateAbout: theaxis, byAngle: -Float(a2))
                lineNode.transform = lineNode.transform * simd_float4x4(scale: SIMD3<Float>(Float(lineSize), Float(length), Float(lineSize)))
                lineNode.color = cc
                nodes.append(lineNode)
                
                let ballNode1 = Node(mesh: LineBallMesh)
                ballNode1.transform = simd_float4x4(translate: SIMD3<Float>(Float(x0), Float(y0), Float(z0)))
                theaxis = SIMD3<Float>(1, 0, 0)
                ballNode1.transform = ballNode1.transform * simd_float4x4(rotateAbout: theaxis, byAngle: .pi/2)
                theaxis = SIMD3<Float>(0, 1, 0)
                ballNode1.transform = ballNode1.transform * simd_float4x4(rotateAbout: theaxis, byAngle: Float(a1))
                theaxis = SIMD3<Float>(0, 0, 1)
                ballNode1.transform = ballNode1.transform * simd_float4x4(rotateAbout: theaxis, byAngle: -Float(a2))
                ballNode1.transform = ballNode1.transform * simd_float4x4(scale: SIMD3<Float>(Float(lineSize), Float(lineSize), Float(lineSize)))
                ballNode1.color = cc
                nodes.append(ballNode1)
                
                let ballNode2 = Node(mesh: LineBallMesh)
                ballNode2.transform = simd_float4x4(translate: SIMD3<Float>(Float(x1), Float(y1), Float(z1)))
                theaxis = SIMD3<Float>(1, 0, 0)
                ballNode2.transform = ballNode2.transform * simd_float4x4(rotateAbout: theaxis, byAngle: .pi/2)
                theaxis = SIMD3<Float>(0, 1, 0)
                ballNode2.transform = ballNode2.transform * simd_float4x4(rotateAbout: theaxis, byAngle: Float(a1))
                theaxis = SIMD3<Float>(0, 0, 1)
                ballNode2.transform = ballNode2.transform * simd_float4x4(rotateAbout: theaxis, byAngle: -Float(a2))
                theaxis = SIMD3<Float>(1, 0, 0)
                ballNode2.transform = ballNode2.transform * simd_float4x4(rotateAbout: theaxis, byAngle: .pi)
                ballNode2.transform = ballNode2.transform * simd_float4x4(scale: SIMD3<Float>(Float(lineSize), Float(lineSize), Float(lineSize)))
                ballNode2.color = cc
                nodes.append(ballNode2)
            }
            last_point = next_point // Update last position every time
        }
        
        // Add Light Sources
        let ambientLight = Light()
        ambientLight.type = .ambient
        ambientLight.intensity = 0.1
        let sunLight = Light()
        sunLight.type = .directional
        sunLight.direction = normalize(SIMD3<Float>(1, 1, 1))
        lights = [ambientLight, sunLight]
    } // End makeScene
    
    
    //###########################################################################################//
    //### DRAW THE VIEW
    //###########################################################################################//
    
    func draw(in metalView: MTKView) {
        frameSemaphore.wait()
        
        // Pause view after this update if no ongoing movement
        if cursorDeltaX == 0  && cursorDeltaZ == 0 { metalView.isPaused = true }
        else { metalView.isPaused = false }
        
        currentWidth = metalView.drawableSize.width
        currentHeight = metalView.drawableSize.height
        
        // Buffer Check - debugging - also set MaxOutstandingFrameCount = 1
        //currentConstantBufferOffset = 0;
        
        updateLightConstants()
        updateFrameConstants()
        updateNodeConstants()
        
        // Buffer Check - debugging
        //print("Buffer offest = ", currentConstantBufferOffset)
        
        guard let renderPassDescriptor = metalView.currentRenderPassDescriptor else { return }
        
        //############################################################################
        // ### Make a double width offscreen texture for RG and RB stereo modes
        //############################################################################
        if stereoMode == .RG || stereoMode == .RB {
            let offScreenTextureDescriptor = MTLTextureDescriptor()
            offScreenTextureDescriptor.width = Int(currentWidth * 2)
            offScreenTextureDescriptor.height = Int(currentHeight)
            offScreenTextureDescriptor.textureType = .type2D
            offScreenTextureDescriptor.pixelFormat = .bgra8Unorm_srgb
            offScreenTextureDescriptor.storageMode = .private
            offScreenTextureDescriptor.usage = [.renderTarget, .shaderRead]
            offScreenTexture = Renderer.device!.makeTexture(descriptor: offScreenTextureDescriptor)
            
            let offScreenDepthTextureDescriptor = MTLTextureDescriptor()
            offScreenDepthTextureDescriptor.width = Int(currentWidth * 2)
            offScreenDepthTextureDescriptor.height = Int(currentHeight)
            offScreenDepthTextureDescriptor.textureType = .type2D
            offScreenDepthTextureDescriptor.pixelFormat = .depth32Float
            offScreenDepthTextureDescriptor.storageMode = .private
            offScreenDepthTextureDescriptor.usage = .renderTarget
            offScreenDepthTexture = Renderer.device!.makeTexture(descriptor: offScreenDepthTextureDescriptor)
            
            renderPassDescriptor.colorAttachments[0].texture = offScreenTexture
            renderPassDescriptor.colorAttachments[0].loadAction = .clear
            renderPassDescriptor.colorAttachments[0].clearColor = MTLClearColor(red: 0, green: 0, blue: 0, alpha: 1)
            renderPassDescriptor.colorAttachments[0].storeAction = .store
            
            renderPassDescriptor.depthAttachment.texture = offScreenDepthTexture
            renderPassDescriptor.depthAttachment.loadAction = .clear
            renderPassDescriptor.depthAttachment.storeAction = .dontCare
            
            renderPassDescriptor.renderTargetWidth = Int(currentWidth * 2)
            renderPassDescriptor.renderTargetHeight = Int(currentHeight)
        }
        //############################################################################
        //############################################################################
        
        guard let commandBuffer = Renderer.commandQueue.makeCommandBuffer() else { return }
        
        let renderCommandEncoder = commandBuffer.makeRenderCommandEncoder(descriptor: renderPassDescriptor)!
        renderCommandEncoder.setRenderPipelineState(renderPipelineState)
        renderCommandEncoder.setDepthStencilState(depthStencilState)
        renderCommandEncoder.setFrontFacing(.counterClockwise)
        renderCommandEncoder.setCullMode(.back)
        renderCommandEncoder.setVertexBuffer(constantBuffer, offset: frameConstantsOffset, index: 3)
        renderCommandEncoder.setFragmentBuffer(constantBuffer, offset: frameConstantsOffset, index: 3)
        renderCommandEncoder.setFragmentBuffer(constantBuffer, offset: lightConstantsOffset, index: 4)
        
        // Single view for mono - Two views for stereo modes
        var  vrViewports: [MTLViewport] = []
        if stereoMode == .dual { // Dual Stereo
            vrViewports.append(MTLViewport(originX: 0, originY: 0, width: currentWidth/2, height: currentHeight, znear: 0.0, zfar: 1.0))
            vrViewports.append(MTLViewport(originX: currentWidth/2, originY: 0, width: currentWidth/2, height: currentHeight, znear: 0.0, zfar: 1.0))
        }
        else if stereoMode == .RG || stereoMode == .RB { // RG or RB Stereo
            vrViewports.append(MTLViewport(originX: 0, originY: 0, width: currentWidth, height: currentHeight, znear: 0.0, zfar: 1.0))
            vrViewports.append(MTLViewport(originX: currentWidth, originY: 0, width: currentWidth, height: currentHeight, znear: 0.0, zfar: 1.0))
        }
        else { // Mono
            vrViewports.append(MTLViewport(originX: 0, originY: 0, width: currentWidth, height: currentHeight, znear: 0.0, zfar: 1.0))
        }
        renderCommandEncoder.setViewports(vrViewports)
        
        // Contruct the renderCommandEncoder - This section takes most of the time
        for (objectIndex, node) in nodes.enumerated() {
            guard let mesh = node.mesh else { continue }
            renderCommandEncoder.setVertexBuffer(constantBuffer, offset: nodeConstantsOffsets[objectIndex], index: 2)
            renderCommandEncoder.setVertexBuffer(mesh.vertexBuffers[0].buffer, offset: mesh.vertexBuffers[0].offset, index: 0)
            var numberOfViews = 1
            if stereoMode != .mono { numberOfViews = 2 }
            let indexBuffer = mesh.submeshes[0].indexBuffer
            renderCommandEncoder.drawIndexedPrimitives(
                type: mesh.submeshes[0].primitiveType,
                indexCount: mesh.submeshes[0].indexCount,
                indexType: mesh.submeshes[0].indexType,
                indexBuffer: indexBuffer.buffer,
                indexBufferOffset: indexBuffer.offset,
                instanceCount: numberOfViews
            )
        }
        renderCommandEncoder.endEncoding()
        
        
        //############################################################################
        // ### Copy offscreen texture to screen for RG and RB stereo modes
        //############################################################################
        if stereoMode == .RG || stereoMode == .RB {
            stereoModePointer = stereoMode.rawValue
            guard let renderPassDescriptor2 = metalView.currentRenderPassDescriptor else { return }
            let renderCommandEncoder2 = commandBuffer.makeRenderCommandEncoder(descriptor: renderPassDescriptor2)!
            renderCommandEncoder2.setRenderPipelineState(renderPipelineState2)
            renderCommandEncoder2.setVertexBytes(quadVertices, length: quadVertices.count * MemoryLayout<QuadTextureVertex>.stride, index: 0)
            renderCommandEncoder2.setVertexBytes(&stereoModePointer, length: MemoryLayout<Int>.stride, index: 1)
            renderCommandEncoder2.setFragmentTexture(offScreenTexture, index: 0)
            renderCommandEncoder2.drawPrimitives(type: .triangle, vertexStart: 0, vertexCount: 6)
            renderCommandEncoder2.endEncoding()
        }
        //############################################################################
        //############################################################################
        
        
        // Submit the drawing commands and present on screen
        commandBuffer.present(metalView.currentDrawable!)
        commandBuffer.addCompletedHandler { [weak self] _ in self?.frameSemaphore.signal() }
        commandBuffer.commit()
        
        // Update FPS display after more than 0.8 seconds have passed since the last draw command
        fpsCount += 1
        let currentTime = CFAbsoluteTimeGetCurrent()
        let passed = currentTime - fpsTime
        if passed > 0.8 {
            updateFPS(String(Int(1.0/passed * Double(fpsCount) + 0.5)) + " FPS")
            fpsTime = currentTime
            fpsCount = 0
        }
    }
    
    
    //###########################################################################################//
    //### UPDATE LIGHT CONSTANTS
    //###########################################################################################//
    
    func updateLightConstants() {
        let layout = MemoryLayout<LightConstants>.self
        lightConstantsOffset = allocateConstantStorage(size: layout.stride * lights.count, alignment: layout.stride)
        let lightsBufferPointer = constantBuffer.contents().advanced(by: lightConstantsOffset).assumingMemoryBound(to: LightConstants.self)
        for (lightIndex, light) in lights.enumerated() {
            lightsBufferPointer[lightIndex] = LightConstants(intensity: light.color * light.intensity, direction: light.direction, type: light.type.rawValue)
        }
    }
    
    
    //###########################################################################################//
    //### UPDATE FRAME CONSTANTS
    //###########################################################################################//
    
    func updateFrameConstants() {
        var aspectRatio = currentWidth / currentHeight // mono RG RB
        if stereoMode == .dual { aspectRatio = currentWidth / currentHeight / 2 } // Dual
        let projectionMatrix = simd_float4x4(perspectiveProjectionFoVY: .pi / 10, aspectRatio: Float(aspectRatio), near: 0.01,far: 1000)
        var constants = FrameConstants(projectionMatrix: projectionMatrix, lightCount: UInt32(lights.count), stereoAngle: Float(stereoAngle), stereoType: Int(stereoMode.rawValue))
        let layout = MemoryLayout<FrameConstants>.self
        frameConstantsOffset = allocateConstantStorage(size: layout.size, alignment: layout.stride)
        let constantsPointer = constantBuffer.contents().advanced(by: frameConstantsOffset)
        constantsPointer.copyMemory(from: &constants, byteCount: layout.size)
    }
    
    
    //###########################################################################################//
    //### UPDATE NODE CONSTANTS
    //###########################################################################################//
    
    func updateNodeConstants() {
        let cameraPosition = SIMD3<Float>(0, 0, Float(zoomLevel*zoomFactor))
        let viewMatrix = simd_float4x4(translate: -cameraPosition)
        let eyeaxis = SIMD3<Float>(0, 1, 0)
        let eyetransform = simd_float4x4(rotateAbout: eyeaxis, byAngle: -.pi / 180 * Float(stereoAngle))
        let viewMatrix1 = viewMatrix * eyetransform
        var rotation3 = simd_float4x4(SIMD4<Float>(1, 0, 0, 0), SIMD4<Float>(0, 1, 0, 0), SIMD4<Float>(0, 0, 1, 0), SIMD4<Float>(0, 0, 0, 1))
        
        // fix rotation angle by looking at time since last here
        var cursorDeltaXfixed = cursorDeltaX
        var cursorDeltaZfixed = cursorDeltaZ
        let passedTime = (CFAbsoluteTimeGetCurrent() - frameTime) * 120
        frameTime = CFAbsoluteTimeGetCurrent()

        if cursorDeltaXfixed != 0 || cursorDeltaZfixed != 0 {
            cursorDeltaXfixed = cursorDeltaXfixed * passedTime
            cursorDeltaZfixed = cursorDeltaZfixed * passedTime
        }
        
        if !trackball {
            /// Turntable Rotation
            absoluteAngleZ -= cursorDeltaXfixed / 50
            absoluteAngleX += cursorDeltaZfixed / 50
            if absoluteAngleX > .pi/2 { absoluteAngleX = .pi/2}
            if absoluteAngleX < -.pi/2 { absoluteAngleX = -.pi/2}
            cursorDeltaZ = 0
            let rotationAxis2 = SIMD3<Float>(-1, 0, 0)
            let rotation2 = simd_float4x4(rotateAbout: rotationAxis2, byAngle: Float(absoluteAngleX))
            let rotationAxis1 = SIMD3<Float>(0, 1, 0)
            let rotation1 = simd_float4x4(rotateAbout: rotationAxis1, byAngle: Float(absoluteAngleZ))
            rotation3 = rotation2 * rotation1
        }
        
        else {
            /// Trackball Rotation
            if (absoluteAngleZ != 0 || absoluteAngleX != 0) {
                let rotationAxis2 = SIMD3<Float>(-1, 0, 0)
                let rotation2 = simd_float4x4(rotateAbout: rotationAxis2, byAngle: Float(absoluteAngleX))
                let rotationAxis1 = SIMD3<Float>(0, 1, 0)
                let rotation1 = simd_float4x4(rotateAbout: rotationAxis1, byAngle: Float(absoluteAngleZ))
                // Rotate by the ZX rotation so it stays in same place and then zero ZX out
                rotation3 = rotation2 * rotation1
                trackballCurrent = rotation3 * trackballCurrent
                absoluteAngleZ = 0
                absoluteAngleX = 0
            }
            let trackAngleZ = -cursorDeltaXfixed / 50
            let trackAngleX = cursorDeltaZfixed / 50
            let rotationAxis2 = SIMD3<Float>(-1, 0, 0)
            let rotation2 = simd_float4x4(rotateAbout: rotationAxis2, byAngle: Float(trackAngleX))
            let rotationAxis1 = SIMD3<Float>(0, 1, 0)
            let rotation1 = simd_float4x4(rotateAbout: rotationAxis1, byAngle: Float(trackAngleZ))
            trackballCurrent = rotation2 * rotation1 * trackballCurrent
        }
        
        if cursorDeltaX != 0 || cursorDeltaZ != 0 {
            cursorDeltaXprevious = cursorDeltaX
            cursorDeltaZprevious = cursorDeltaZ
        }
        
        nodeConstantsOffsets.removeAll()
        for node in nodes {
            let transform = viewMatrix * rotation3 * trackballCurrent * node.transform
            let transform1 = viewMatrix1 * rotation3 * trackballCurrent * node.transform
            var constants = NodeConstants(modelViewMatrix: transform, modelViewMatrix1: transform1, color: node.color)
            let layout = MemoryLayout<NodeConstants>.self
            let offset = allocateConstantStorage(size: layout.size, alignment: layout.stride)
            let constantsPointer = constantBuffer.contents().advanced(by: offset)
            constantsPointer.copyMemory(from: &constants, byteCount: layout.size)
            nodeConstantsOffsets.append(offset)
        }
    }
    
    
    //###########################################################################################//
    //### MAKE PIPELINES
    //###########################################################################################//
    
    func makePipeline(metalView: MTKView) {
        guard let library = Renderer.device.makeDefaultLibrary() else { fatalError("Metal library Error") }
        
        // Main drawing pipeline
        let renderPipelineDescriptor = MTLRenderPipelineDescriptor()
        renderPipelineDescriptor.label = "Draw Scene"
        renderPipelineDescriptor.vertexDescriptor = vertexDescriptor
        renderPipelineDescriptor.colorAttachments[0].pixelFormat = metalView.colorPixelFormat
        renderPipelineDescriptor.depthAttachmentPixelFormat = metalView.depthStencilPixelFormat
        renderPipelineDescriptor.vertexFunction = library.makeFunction(name: "vertex_main")!
        renderPipelineDescriptor.fragmentFunction = library.makeFunction(name: "fragment_main")!
        //renderPipelineDescriptor.vertexBuffers[0].mutability = .immutable
        do { renderPipelineState = try Renderer.device.makeRenderPipelineState(descriptor: renderPipelineDescriptor) }
        catch { fatalError("Draw Scene Error: \(error)") }
        
        // Copy offscreen texture to screen pipeline
        let renderPipelineDescriptor2 = MTLRenderPipelineDescriptor()
        renderPipelineDescriptor.label = "Draw to Screen"
        renderPipelineDescriptor2.vertexDescriptor = vertexDescriptor
        renderPipelineDescriptor2.colorAttachments[0].pixelFormat = metalView.colorPixelFormat
        renderPipelineDescriptor2.depthAttachmentPixelFormat = metalView.depthStencilPixelFormat
        renderPipelineDescriptor2.vertexFunction = library.makeFunction(name: "textureVertexShader")!
        renderPipelineDescriptor2.fragmentFunction = library.makeFunction(name: "textureFragmentShader")!
        //renderPipelineDescriptor2.vertexBuffers[0].mutability = .immutable
        do { renderPipelineState2 = try Renderer.device.makeRenderPipelineState(descriptor: renderPipelineDescriptor2) }
        catch { fatalError("Draw Screen Error: \(error)") }
        
        // Depth render state
        let depthStencilDescriptor = MTLDepthStencilDescriptor()
        depthStencilDescriptor.depthCompareFunction = .less
        depthStencilDescriptor.isDepthWriteEnabled = true
        depthStencilState = Renderer.device.makeDepthStencilState(descriptor: depthStencilDescriptor)!
    }
    
    
    //###########################################################################################//
    //### ALLOCATE CONSTANT STORAGE
    //###########################################################################################//
    
    func allocateConstantStorage(size: Int, alignment: Int) -> Int {
        let effectiveAlignment = lcm(alignment, MinBufferAlignment)
        var allocationOffset = align(currentConstantBufferOffset, upTo: effectiveAlignment)
        if (allocationOffset + size >= MaxConstantsSize) { allocationOffset = 0 }
        currentConstantBufferOffset = allocationOffset + size
        return allocationOffset
    }
    
    
    //###########################################################################################//
    //### WINDOW SIZE CHANGED
    //###########################################################################################//
    
    func mtkView(_ metalView: MTKView, drawableSizeWillChange size: CGSize) {
        makeScene(metalView: metalView)
        metalView.isPaused = false
    }
    
    
    //##################################################################################//
    //### Handle User Interface Events
    //##################################################################################//
    
    func dragUpdate(currentMousePoint: CGPoint, dragEnded: Bool, metalView: MTKView) {
        let currentTime = CFAbsoluteTimeGetCurrent()
        if dragEnded {
            // End of drag event - continue rotation
            let mouseTimePassed = currentTime - mouseTime
            if mouseTimePassed < 0.25 {
                // Only continue rotation if not stationary for more than 0.25 seconds
                var xSum = 0.0
                var ySum = 0.0
                let endTime = moveHistory[moveHistory.count-1].time
                var startTime = endTime
                for h in moveHistory.reversed() {
                    if (currentTime - h.time) < 0.5 {
                        // Average movement over the last half a second
                        xSum += h.x
                        ySum += h.y
                        startTime = h.time
                    }
                    else { break }
                }
                let timePassed = (endTime - startTime) * 200
                if timePassed != 0 {
                    cursorDeltaX = xSum / timePassed / 2
                    cursorDeltaZ = ySum / timePassed / 2
                }
                metalView.isPaused = false
            }
            moveHistory = []
            return
        }
        // Dragging is actively happening
        mouseTime = currentTime
        if currentMousePoint == CGPoint(x: 0, y: 0) { previousMousePoint = currentMousePoint }
        cursorDeltaX = 0
        cursorDeltaZ = 0
        let xDiff = currentMousePoint.x - previousMousePoint.x
        let yDiff = currentMousePoint.y - previousMousePoint.y
        absoluteAngleZ -= xDiff / 200
        absoluteAngleX += yDiff / 200
        previousMousePoint = currentMousePoint
        let tempHistory = MoveHistory(time: currentTime, x: xDiff, y: yDiff)
        moveHistory.append(tempHistory)
        metalView.isPaused = false
    }
    
    
    func zoomUpdate(zoomChange: Double, zoomEnded: Bool, metalView: MTKView) {
        if zoomEnded {
            zoomLevel = zoomLevel * zoomFactor
            zoomFactor = 1.0
        }
        else { zoomFactor = 1.0 / zoomChange }
        makeScene(metalView: metalView)
        metalView.isPaused = false
    }
    
    
    func resetAngles(metalView: MTKView) {
        metalView.isPaused = true
        absoluteAngleX = 0
        absoluteAngleZ = 0
        cursorDeltaX = 0
        cursorDeltaZ = 0
        zoomLevel = 7
        trackballCurrent = simd_float4x4(SIMD4<Float>(1, 0, 0, 0), SIMD4<Float>(0, 1, 0, 0), SIMD4<Float>(0, 0, 1, 0), SIMD4<Float>(0, 0, 0, 1))
        metalView.isPaused = false
    }
    
    
    func factoryReset(metalView: MTKView) {
        metalView.isPaused = true
        zoomLevel = 7
        ballSize = 0.04
        lineSize = 0.02
        upDirection = "Z"
        invertObject = false
        handedness = false
        absoluteAngleX = 0
        absoluteAngleZ = 0
        centerObject = false
        stereoMode = .mono
        stereoAngle = 3.2
        wireFrame = false
        trackball = false
        backgroundColor = 0
        setBackgroundColor(backgroundColor, metalView: metalView)
        cursorDeltaX = 0
        cursorDeltaZ = 0
        trackballCurrent = simd_float4x4(SIMD4<Float>(1, 0, 0, 0), SIMD4<Float>(0, 1, 0, 0), SIMD4<Float>(0, 0, 1, 0), SIMD4<Float>(0, 0, 0, 1))
        makeScene(metalView: metalView)
        metalView.isPaused = false
    }
    
    
    func saveDefaults(metalView: MTKView) {
        metalView.isPaused = true
        @AppStorage("DefaultBallSize") var DefaultBallSize = 0.04
        DefaultBallSize = Double(ballSize)
        @AppStorage("DefaultLineSize") var DefaultLineSize = 0.02
        DefaultLineSize = Double(lineSize)
        @AppStorage("DefaultUpDirection") var DefaultUpDirection = "Z"
        DefaultUpDirection = String(upDirection)
        @AppStorage("DefaultInvertObject") var DefaultInvertObject = false
        DefaultInvertObject = invertObject
        @AppStorage("DefaultHandedness") var DefaultHandedness = false
        DefaultHandedness = handedness
        @AppStorage("DefaultCenterObject") var DefaultCenterObject = false
        DefaultCenterObject = centerObject
        @AppStorage("DefaultBackgroundColor") var DefaultBackgroundColor = 0
        DefaultBackgroundColor = backgroundColor
        @AppStorage("DefaultStereoMode") var DefaultStereoMode = 0
        DefaultStereoMode = stereoMode.rawValue
        @AppStorage("DefaultStereoAngle") var DefaultStereoAngle = 3.2
        DefaultStereoAngle = stereoAngle
        @AppStorage("DefaultTrackball") var DefaultTrackball = false
        DefaultTrackball = trackball
        @AppStorage("DefaultWireFrame") var DefaultWireFrame = false
        DefaultWireFrame = wireFrame
        metalView.isPaused = false
    }
    
    
    //###########################################################################################//
    //### Set Background Color
    //###########################################################################################//
    
    func setBackgroundColor(_ theColor: Int, metalView: MTKView) {
        if theColor == 0 { metalView.clearColor = MTLClearColor(red: 0.0, green: 0.0, blue: 0.0, alpha: 1.0) }
        else if theColor == 1 { metalView.clearColor = MTLClearColor(red: 1.0, green: 1.0, blue: 1.0, alpha: 1.0) }
        else { metalView.clearColor = MTLClearColor(red: 0.3, green: 0.3, blue: 0.3, alpha: 1.0) }
    }
    
    
    //##################################################################################//
    //### Parse file text into object data array and create constantBuffer
    //##################################################################################//
    
    func readDataText(inputText: String, metalView: MTKView) {
        let the_lines = inputText.components(separatedBy: ["\n", "\r"]).filter({!$0.isEmpty})
        objectDataArray = []
        var number_balls = 0
        var number_lines = 0
        
        for w in the_lines {
            // Ignore text after #
            let match2 = matches(for: "([^#]*)", in: w)
            if match2[0] != "" {
                // Break a line up into number like things
                let match3 = matches(for: "[0-9Ee\\.+-]+", in: match2[0])
                if match3.count > 0 {
                    var nums = Array<Double>()
                    for k in match3 {
                        let z = Double(k)
                        if z != nil {
                            let zz = z!
                            nums.append(zz)
                        }
                    }
                    if nums.count == 2 { nums.append(0) }
                    if nums.count == 3 { nums.append(-7) }
                    if nums.count >= 4 {
                        var y = Array<Double>()
                        if nums[3] > 7.0 { nums[3] = 7.0}
                        if nums[3] < -7.0 { nums[3] = -7}
                        if nums[3] < 0 { number_balls += 1 }
                        if nums[3] > 0 { number_lines += 1 }
                        for yy in nums[0...3] { y.append(yy) }
                        objectDataArray.append(y)
                    }
                }
            }
        }
        metalView.isPaused = true
        
        // Create constantBuffer by calculating its size
        MaxConstantsSize = (136 + 6912 * number_lines + 2304 * number_balls) * MaxOutstandingFrameCount
        MaxConstantsSize = (Int(MaxConstantsSize / 4096) + 1) * 4096
        constantBuffer = Renderer.device.makeBuffer(length: MaxConstantsSize, options: .storageModeShared)
        constantBuffer.label = "Dynamic Constant Buffer"
        
        makeScene(metalView: metalView)
        metalView.isPaused = false
        
        func matches(for regex: String, in text: String) -> [String] {
            do {
                let regex = try NSRegularExpression(pattern: regex)
                let results = regex.matches(in: text, range: NSRange(text.startIndex..., in: text))
                return results.map { String(text[Range($0.range, in: text)!]) }
            }
            catch let error {
                print("invalid regex: \(error.localizedDescription)")
                return []
            }
        }
    }
}
