
#include <metal_stdlib>

using namespace metal;


enum LightType : uint {
    LightTypeAmbient,
    LightTypeDirectional
};


struct Light {
    vector_float3 intensity; // product of color and intensity
    vector_float3 direction;
    LightType type;
};


struct VertexIn {
    vector_float3 position  [[attribute(0)]];
    vector_float3 normal    [[attribute(1)]];
};


struct VertexOut {
    vector_float4 position [[position]];
    vector_float3 viewPosition;
    vector_float3 normal;
    int color;
    ushort viewport [[ viewport_array_index ]]; // Two views for stereo
};


struct NodeConstants {
    float4x4 modelViewMatrix;
    float4x4 modelViewMatrix1;
    int color;
};


struct FrameConstants {
    float4x4 projectionMatrix;
    uint lightCount;
    float stereoAngle;
    int stereoType;
};


//###############################################################################
// ##############################################################################
// ### Render object
// ##############################################################################
//###############################################################################

vertex VertexOut
vertex_main(VertexIn in [[stage_in]],
            constant NodeConstants &node [[buffer(2)]],
            constant FrameConstants &frame [[buffer(3)]],
            ushort viewSide [[ instance_id ]] ) {
    vector_float4 viewPosition;
    vector_float4 viewNormal;
    if (viewSide != 0) {
        viewPosition = node.modelViewMatrix1 * vector_float4(in.position, 1.0);
        viewNormal = node.modelViewMatrix1 * vector_float4(in.normal, 0.0);
    }
    else {
        viewPosition = node.modelViewMatrix * vector_float4(in.position, 1.0);
        viewNormal = node.modelViewMatrix * vector_float4(in.normal, 0.0);
    }
    int color = node.color;
    VertexOut out;
    out.position = frame.projectionMatrix * viewPosition;
    out.viewPosition = viewPosition.xyz;
    out.normal = viewNormal.xyz;
    out.color = color;
    out.viewport = viewSide;
    return out;
}


fragment vector_float4
fragment_main(VertexOut in [[stage_in]],
              constant NodeConstants &node [[buffer(2)]],
              constant FrameConstants &frame [[buffer(3)]],
              constant Light *lights [[buffer(4)]] ) {
    vector_float4 baseColor = { 0.8, 0.8, 0.8, 1 };        // white
    switch(in.color) {
        case 1: baseColor = { 1, 0, 0, 1 }; break;         // red
        case 2: baseColor = { 0, 1, 0, 1 }; break;         // green
        case 3: baseColor = { 0, 0, 1, 1 }; break;         // blue
        case 4: baseColor = { 0.8, 0.8, 0, 1 }; break;     // yellow
        case 5: baseColor = { 1, 0, 1, 1 }; break;         // purple
        case 6: baseColor = { 0, 1, 1, 1 }; break;         // cyan
        case 7: baseColor = { 0.8, 0.8, 0.8, 1 }; break;   // white
    }
    
    float specularExponent = 50.0;
    
    if (frame.stereoType > 1 && in.viewport == 0) { baseColor = { 1, 0, 0, 1 }; }
    if (frame.stereoType == 2 && in.viewport == 1) { baseColor = { 0, 1, 0, 1 }; }
    if (frame.stereoType == 3 && in.viewport == 1) { baseColor = { 0, 0, 1, 1 }; }
    
    // Change location of sun light source for second stereo image
    // Rotate around Y axis (up screen)
    vector_float3 sunDirection = lights[1].direction;
    float stereoAngle = 0.017453292519943 * frame.stereoAngle; // Degrees to Radians
    if (in.viewport != 0) {
        float x = sunDirection[0];
        float z = sunDirection[2];
        sunDirection[0] = x * cos(-stereoAngle) - z * sin(-stereoAngle);
        sunDirection[2] = x * sin(-stereoAngle) + z * cos(-stereoAngle);
    }
    
    vector_float3 N = normalize(in.normal);
    vector_float3 V = normalize(vector_float3(0) - in.viewPosition);
    vector_float3 litColor { 0 };
    
    for (uint i = 0; i < frame.lightCount; ++i) {
        float ambientFactor = 0;
        float diffuseFactor = 0;
        float specularFactor = 0;
        constant Light &light = lights[i];
        
        switch(light.type) {
            case LightTypeAmbient:
                ambientFactor = 1;
                break;
            case LightTypeDirectional: {
                vector_float3 L = normalize(sunDirection);
                vector_float3 H = normalize(L + V);
                diffuseFactor = saturate(dot(N, L));
                specularFactor = powr(saturate(dot(N, H)), specularExponent);
                break;
            }
        }
        
        litColor += (ambientFactor + diffuseFactor) * light.intensity * baseColor.rgb;
        if (frame.stereoType <= 1) { litColor += specularFactor * light.intensity * 0.5; }
    }
    return vector_float4(litColor * baseColor.a, baseColor.a);
}



//###############################################################################
// ##############################################################################
// ### Copy offscreen texture to the view for RG and RB stereo modes
// ##############################################################################
//###############################################################################

typedef struct {
    vector_float2 position;
    vector_float2 texcoord;
} QuadTextureVertex;


typedef struct {
    vector_float4 position [[position]];
    vector_float2 texcoord;
    int stereoType;
} TexturePipelineRasterizerData;


//###############################################################################
// Vertex shader passes texture coordinates through to the rasterizer
//###############################################################################

vertex TexturePipelineRasterizerData
textureVertexShader(const uint vertexID [[ vertex_id ]],
                    const device QuadTextureVertex *vertices [[ buffer(0) ]],
                    constant int &stereoType [[ buffer(1) ]] ) {
    TexturePipelineRasterizerData out;
    out.position = vector_float4(0.0, 0.0, 0.0, 1.0);
    out.position.x = vertices[vertexID].position.x;
    out.position.y = vertices[vertexID].position.y;
    out.texcoord = vertices[vertexID].texcoord;
    out.stereoType = stereoType;
    return out;
}


//###############################################################################
// Fragment shader that samples a texture and outputs the sampled color
//###############################################################################

fragment vector_float4
textureFragmentShader(TexturePipelineRasterizerData in [[stage_in]],
                      texture2d<float> texture [[texture(0)]] ) {
    sampler simpleSampler;
    vector_float4 colorSample;
    
    vector_float2 left = in.texcoord;
    vector_float2 right = in.texcoord;
    left.x = left.x / 2;
    right.x = right.x / 2 + 0.5;
    vector_float4 leftcolor = texture.sample(simpleSampler, left);
    vector_float4 rightcolor = texture.sample(simpleSampler, right);
    
    // Fix non-black background maybe
    colorSample = vector_float4(0.0, 0.0, 0.0, 1.0);
    colorSample.r = leftcolor.r;
    if (in.stereoType == 2) { colorSample.g = rightcolor.g; } // RG
    if (in.stereoType == 3) { colorSample.b = rightcolor.b; } // RB
    
    // Output full texture for testing
    //colorSample = texture.sample(simpleSampler, in.texcoord);
    
    return colorSample;
}
